
from omega   import OMEGA as omg
from waveamp import SNR
from numpy   import pi,sin,cos,sqrt,log,exp,power

import os
import numpy as np
import scipy


def prange(fs, psds, m=1.4, thd=0, frac=0, dc=1, smin=4, opt=None, **args):
    '''
    Get a range from a list of frequency (fs) and PSD (psds)
    
    '''
    if not isinstance(psds, list) and not isinstance(fs, list):
        fs   = [fs]
        psds = [psds]

    if len(fs) != len(psds):
        raise ValueError('Mismatch between fs(%d) and psds(%d)'\
                             % (len(fs), len(psds)))

    # Get a range with Sky Confidence Area (CA) below the threshold (thd)
    if opt == 'CA' or opt == 'ca' or opt == 'area':
        ca = thd if thd > 0 else 10
        return v2r(avol(fs, psds, m=m, thd=ca, dc=dc, smin=smin, **args))

    # Get a horizon 
    elif opt == 'hor' or opt == 'horz':
        SNR = thd if thd > 0 else 8 if len(fs) == 1 else 12
        szh = szhor(fs, psds, m=m, thd=SNR, **args)
        return omg.LumD(szh) if opt == 'hor' else szh

    # Get a responce distance
    elif opt == 'frac' or opt == 'fracz':
        SNR = thd if thd > 0 else 8 if len(fs) == 1 else 12
        szh = zfrac(fs, psds, m=m, thd=SNR, frac=frac, **args)
        return omg.LumD(szh) if opt == 'frac' else szh

    # Get a range with SNR larger than the threshold (thd)
    else:
        SNR = thd if thd > 0 else 8 if len(fs) == 1 else 12
        return v2r(svol(fs, psds, m=m, thd=SNR, dc=dc, smin=smin, opt=opt,\
                            **args))

def frange(psds, m=1.4, thd=0, dc=1, opt=None, **args):
    '''
    Get a range with a list of PSD files
    '''
    fs,ps = loadPSDs(psds)
    return prange(fs, ps, m=m, thd=thd, dc=dc, opt=opt, **args)

def loadNet(gen, isK= False):
    '''
    Load PSD files corresponding a specified detector network
    Available networks are: O3, O4low, O4(O4high), O5low, O5high(O5), 3G
    Default KAGRA sensitivity is added only if isK is True for O3 and O4
    '''
    dn = os.path.dirname(__file__)
    dn = dn[:dn.rfind('/')]
    dn += '/psd/'
    fn = []

    if gen == 'O3':
        fn = ['aligo_120Mpc.txt', 'aligo_120Mpc.txt', 'avirgo_60Mpc.txt']
        if isK:
            fn.append('psd/kagra_25Mpc.txt')

    elif gen == 'O4low':
        fn = ['aligo_O4low.txt', 'aligo_O4low.txt', 'avirgo_O4low.txt']
        if isK:
            fn.append('kagra_25Mpc.txt')

    elif gen == 'O4high' or gen == 'O4':
        fn = ['aligo_O4high.txt', 'aligo_O4high.txt', 'avirgo_O4high.txt']
        if isK:
            fn.append('kagra_80Mpc.txt')

    elif gen == 'O5low':
        fn = ['AplusDesign.txt', 'AplusDesign.txt', 'avirgo_O5low.txt']

    elif gen == 'O5high' or gen == 'O5':
        fn = ['AplusDesign.txt', 'AplusDesign.txt', 'avirgo_O5high.txt']

    elif gen == '3G':
        fn = ['CE-W.txt', 'ET-D.txt']

    else:
        raise ValueError('Unsuppoted generation: '+gen)

    for i in range(len(fn)):
        fn[i] = dn+fn[i]

    return loadPSDs(fn)

def loadPSDs(fns, PSD= True):
    '''
    Load PSD files and return list of frequency and PSD data
    '''
    if not isinstance(fns, list):
        return loadPSD(fns, PSD)

    fs = []
    ps = []
    for fn in fns:
        f, p = loadPSD(fn)
        fs.append(f)
        ps.append(p)

    return fs,ps

def loadPSD(fn, PSD= True):
    '''
    Load PSD file and return frequency and PSD
    '''
    fh = np.loadtxt(fn)
    f  = fh[:,0]
    h  = fh[:,1]
    if PSD and np.min(h) > 1e-30:
        h *= h
    return f,h

def svol(fs, psds, m=1.4, thd=8, dc=1, smin=4, opt=None, **args):
    '''
    Detector redshift-corrected comoving volume

    For the given detector noise PSD (fs,psds) return
    the redshift-corrected, comoving volume in Mpc^3 within which all
    sources would have SNR greater than threshold (thd)

    '''
    zh   = szhor(fs, psds, m=m, thd=thd, **args)
    x,w  = scipy.special.orthogonal.p_roots(20)
    x    = np.real(x)
    zz   = 0.5*zh*(x+1)
    nd   = len(fs)
    df   = fduty(nd, dc)[0] if dc < 1 else np.ones(1)
    frac = []

    if (len(fs) == 1 and opt is not 'resp') or \
       (len(fs) >  1 and opt is     'opt'):
        ss = np.array([SNRz(fs, psds, z, m=m, **args) for z in zz])
        frac.append(np.array([angav(snr/thd) for snr in ss]))
    else:
        resp = dresp(len(fs))
        n0   = len(resp[0])
        frac = [[] for i in range(len(df))]

        for z in zz:
            ss = SNRz(fs, psds, z, m=m, resp=resp, smin=smin, dc=dc, **args)
            for j in np.arange(len(df)):
                sj = ss[j*n0:j*n0+n0]
                sj.sort()
                if sj[0] > thd:
                    frac[j] = np.append(frac[j], 1)
                elif sj[-1] < thd:
                    frac[j] = np.append(frac[j], 0)
                else:
                    x = np.linspace(1, 0, len(sj), endpoint=False)
                    frac[j] = np.append(frac[j],\
                                        scipy.interpolate.interp1d(sj, x)(thd))

    # comoving volume density, dVc/dz/(1+z)
    dVz = np.array([omg.CVD(z) for z in zz])

    vv = 0
    for j in np.arange(len(df)):
        v = 0.5*zh*sum(w*dVz*frac[j])
        vv += v*df[j]

    # Sensitivity volume in Mpc^3
    return vv

def avol(fs, psds, m=1.4, thd=10, sthd=12, dc=1, smin=4, **args):
    '''
    Detector redshift-corrected comoving volume

    For the given detector noise PSD (fs,psds) return
    the redshift-corrected, comoving volume in Mpc^3 within which all
    sources would have sky localization are better than threshold (thd) deg^2

    '''
    zh   = szhor(fs, psds, m=m, **args)
    x, w = scipy.special.orthogonal.p_roots(25)
    x    = np.real(x)
    zz   = 0.5*zh*(x+1)
    nd   = len(fs)
    resp = dresp(nd)
    n0   = len(resp[0])
    df   = fduty(nd, dc, nmin=3)[0] if dc < 1 else np.ones(1)
    frac = [[] for i in range(len(df))]
    fms  = []
    cal  = []

    nctj = 5
    npsi = 5
    ra   = np.array([])
    dec  = np.array([])
    ctj  = np.array([])
    psi  = np.array([])
    n    = len(RA)
    for i in np.arange(0, nctj):
        for j in np.arange(0, npsi):
            ra  = np.append(ra,  RA)
            dec = np.append(dec, DEC)
            ctj = np.append(ctj, np.ones(n)*(i+0.5)/nctj)
            psi = np.append(psi, np.ones(n)*(j+0.5)/npsi*pi)

    ap0 = apat(psi, ctj, gmst=0, N=nd, ra=ra, dec=dec)
    dc0 = []
    for i in range(0, nd):
        dc0.append(detcc(i, ra, dec))

    for z in zz:
        cas = arngzn(fs, psds, z, m=m, dc=dc, ap0=ap0, dc0=dc0, **args)
        sss = SNRz  (fs, psds, z, m=m, dc=dc, resp=resp)

        for j in np.arange(len(df)):
            ca = cas[j*n0:j*n0+n0]
            ss = sss[j*n0:j*n0+n0]
            ii = np.where(ss > sthd)[0]

            fm = 1.*len(ii)/len(ss)
            ca = ca[ii]
            ca.sort()

            fms.append(fm)
            cal.append(len(ca))

            if len(ca) == 0 or ca[0] > thd:
                frac[j] = np.append(frac[j], 0)
            elif ca[-1] < thd:
                frac[j] = np.append(frac[j], fm)
            else:
                x = np.linspace(0, fm, len(ca), endpoint=False)
                frac[j] = np.append(frac[j],\
                                    scipy.interpolate.interp1d(ca, x)(thd))

    # comoving volume density, dVc/dz/(1+z)
    dVz = np.array([omg.CVD(z) for z in zz])

    vv = 0
    for j in np.arange(len(df)):
        v = 0.5*zh*sum(w*dVz*frac[j])
        vv += v*df[j]

    # Sensitivity volume in Mpc^3
    return vv

def szhor(fs, psds, m=1.4, thd=8, zmin=1e-8, zmax=10, **args):
    '''
    Get the horizon redshift where SNR is thredhold(thd)
    '''
    def SNRzs(z):
        return SNRz(fs, psds, z, m=m, **args)-thd

    return scipy.optimize.brentq(SNRzs, zmin, zmax)


def zfrac(fs, psds, m=1.4, thd=8, frac=0.1, zmin=1e-8, zmax=10, **args):
    '''
    For the given detector noise PSD and waveform parameters return
    the distance in Mpc at which the specified fraction of sources
    would be detected (SNR >= CANONICAL_SNR) if they were all placed
    at that distance.  Assumes a uniform distribution of sources.

    Remaining keyword arguments are interpreted as waveform generation
    parameters (see waveform.gen_waveform()), or a pre-generated
    fiducial waveform `H` may be provided (see waveform.CBCWaveform).
    '''
    def SNRzf(z):
        snr = SNRz(fs, psds, z, m=m, **args)
        f   = angav(snr/thd)
        return f-frac

    return scipy.optimize.brentq(SNRzf, zmin, zmax)

def azhor(fs, psds, m=1.4, thd=10, cl0=0.9, ap=True,
          zmin=1e-8, zmax=10, **args):
    '''
    Get the horizon redshift where sky localization is thredhold(thd) deg^2
    '''
    def arngzc(z):
        return np.min(areaz(fs, psds, z, m=m, ap=ap, cl=cl0, **args))-thd

    return scipy.optimize.brentq(arngzc, zmin, zmax)

def SNRz(fs, psds, z, m, resp=None, smin=4, nmin=2, dc=1, **args):
    '''
    Get a SNR of inspiral at z from a list of frequency(fs) and PSD(psds)
    '''
    snrs = []
    nd   = len(fs)
    n    = len(resp[0]) if resp is not None else 1
    n0   = n
    for i in np.arange(nd):
        f   = fs  [i]
        psd = psds[i]
        snrs.append(np.ones(n)*\
                    SNR(f, psd, d= omg.LumD(z), m= m*(1+z), **args)**2)

    if dc < 1:
        cp = fduty(nd, dc, nmin=nmin)[1]
        n *= len(cp)

    snr = np.zeros(n)
    ns  = np.zeros(n)
    for i in np.arange(nd):
        rf = resp[i] if resp is not None else np.ones(1)
        if dc < 1:
            for j in np.arange(len(cp)):
                if cp[j][i]:
                    ss = np.sqrt(snrs[i])/rf
                    ii = np.where(ss > smin)[0]
                    snr[j*n0:j*n0+n0] += ss**2
                    ns [j*n0+ii] += 1
        else:
            ii   = np.where(sqrt(snrs[i])/rf > smin)[0]
            snr += snrs[i]/rf**2
            ns[ii] += 1

    if resp is None:
        snr = snr[0]
    else:
        ii  = np.where(ns < nmin)[0]
        snr[ii] = 0

    return sqrt(snr)

from copy import copy

def arngzn(fs, psds, z, m=1.4, dc=1, nctj=5, npsi=5,\
           ap0=None, dc0=None, **args):
    '''
    Sky localization area at z with inclination and polaziation grid
    '''
    nd  = len(fs)
    st0 = stz(fs, psds, z, m)
    cas = []
    st1 = []

    if dc < 1:
        cps = fduty(nd, dc, nmin=3)[1]
        for cp in cps:
            st1.append(copy(st0))
            ii = np.where(cp == 0)[0]
            if len(ii) > 0:
                st1[-1][ii] = 100
    else:
        st1.append(st0[:])

    for st in st1:
        ra  = np.array([])
        dec = np.array([])
        ctj = np.array([])
        psi = np.array([])
        n   = len(RA)

        if ap0 is None or dc0 is None:
            ap0 = True
            for i in np.arange(0, nctj):
                for j in np.arange(0, npsi):
                    ra  = np.append(ra,  RA)
                    dec = np.append(dec, DEC)
                    ctj = np.append(ctj, np.ones(n)*(i+0.5)/nctj)
                    psi = np.append(psi, np.ones(n)*(j+0.5)/npsi*pi)

        ca  = area(st, psi=psi, ctj=ctj, N=nd, ra=ra, dec=dec,\
                       ap=ap0, detc0=dc0, **args)
        cas = np.append(cas, ca)

    return cas

def areaz(fs, psds, z, m=1.4, **args):
    '''
    Sky localization area at z
    '''
    st = stz(fs, psds, z, m)
    return area(st, N= len(st), **args)

def stz(fs, psds, z, m=1.4):
    '''
    Get a list of sigmat(m) at z
    '''
    st = np.ndarray(0)
    nd = len(fs)
    d  = omg.LumD(z)

    for i in np.arange(nd):
        f   = fs  [i]
        psd = psds[i]
        st = np.append(st, sigmat(f, psd, dist= d, m= m*(1+z)))

    return st

def sigmat(f, p, dist=100, m=1.4):
    '''
    Timing resolution of given PSD (f,p)
     for an inspiral signal with mass(m) and distance(dist)
    '''
    return 1/SNR(f, p, dist, m=m, fpw=1)/2/pi


def fduty(nd, dc, nmin = 2):
    '''
    Detector network coincidence pattern (cp) and duty factor (df)
    '''

    cp = []
    if nd == 3:
        cp = [np.ones(3)]
        df =  np.ones(1)*dc**3 # 3/3

    elif nd == 4:
        cp = [np.ones(4)]
        df =  np.ones(1)*dc**4 # 4/4

        if nmin <= 3:
            cp.extend([np.ones(4), np.ones(4), np.ones(4), np.ones(4)])
            cp[1][0] = cp[2][1] = cp[3][2] = cp[4][3] = 0
            df = np.append(df, np.ones(4)*dc**3*(1-dc)) # 3/4

    else:
        raise ValueError('Unsuppoted nd(%d)' % nd)

    if nmin == 2:
        if nd == 3:
            cp.extend([[1,1,0], [1,0,1], [0,1,1]])
            df = np.append(df, np.ones(3)*dc**2*(1-dc))     # 2/3

        elif nd == 4:
            cp.extend([[1,1,0,0], [1,0,1,0], [1,0,0,1],
                       [0,1,1,0], [0,1,0,1], [0,0,1,1]])
            df = np.append(df, np.ones(6)*dc**2*(1-dc)**2)  # 2/4

    return df,cp

def dresp(N=1, nctj=5, npsi=5):
    '''
    Get a list of detector responses
    in uniform RA,DEC,cos(theta_JN) and psi bins
    '''
    ff = [[] for i in range(N)]
    for i in np.arange(nctj):
        for j in np.arange(npsi):
            ctj =       (i+0.5)/nctj
            psi = np.pi*(j+0.5)/npsi
            f   = apat(psi, ctj, 0, N)

            for k in np.arange(N):
                ff[k] = np.append(ff[k], f[k])
    return ff

def v2r(v):
    '''
    Get the radius of a sphere with volume (v)
    '''
    return (3*v/4/np.pi)**(1./3)

def z2v(z, n=20):
    '''
    Get redshift-corrected comoving volume in Mpc^3 at the given redshift (z)
    '''
    x,w = scipy.special.orthogonal.p_roots(n)
    x   = np.real(x)
    zz  = 0.5*z*(x+1)
    dV  = np.array([omg.CVD(z) for z in zz])
    return 0.5*z*sum(w*dV)

def d2v(d, n=20):
    '''
    Get redshift-corrected comoving volume in Mpc^3
     at the given luminosity distance (d/Mpc)
    '''
    return z2v(omg.z(d))

def d2r(d):
    '''
    Approx. function of v2r(d2v(d))
    '''
    u = [-6.9078,  -5.8739,  -4.8401,  -4.3232,  -3.8063,  -3.2894,
         -2.7725,  -2.2556,  -1.7387,  -1.2218,  -0.7049,  -0.1880,
          0.3289,   0.8458,   1.3628,   1.8797,   2.3966,   2.9135,
          3.4304,   3.9473,   4.4642,   4.9811,   5.4980,   6.0149,
          6.5318,   7.0487,   7.5656,   8.0825,   8.5995,   9.1164,
          9.6333,  10.1502,  10.6671,  11.1840,  11.7009,  12.2178,
         12.7347,  13.2516,  13.7685,  14.2854,  14.8023,  15.3192,
         15.8361,  16.3531,  16.8700,  17.3869,  17.9038,  18.4207 ]
    v = [-1.1468,  -1.1473,  -1.1488,  -1.1503,  -1.1528,  -1.1571,
         -1.1642,  -1.1762,  -1.1962,  -1.2297,  -1.2855,  -1.3780,
         -1.5294,  -1.7727,  -2.1520,  -2.7176,  -3.5078,  -4.5117,
         -5.6115,  -6.5243,  -6.7928,  -5.9028,  -3.5757,  -0.1500,
          3.2557,   5.1627,   4.6588,   2.0646,  -1.3211,  -4.0191,
         -4.9445,  -3.7195,  -0.8818,   2.3608,   4.8445,   6.0106,
          5.9498,   5.1053,   3.9500,   2.8212,   1.9163,   1.2789,
          0.8314,   0.5090,   0.2904,   0.1560,   0.0801,   0.0398 ]

    if d < np.exp(u[0]):
        return d*(1+v[0]*1e-3)

    p0 = 8.454
    p1 = 2.692
    x  = np.log(d)
    r  = scipy.special.erf((p0-x)/p1)/2+0.5
    r += 1e-3*scipy.interpolate.interp1d(u, v)(x)
    return r*d


def angav(x):
    '''
    Approx. function of angular averaged single detector response
    '''
    u = [ 0.0000,   0.0930,   0.1861,   0.2791,   0.3721,   0.4652,
          0.5582,   0.6512,   0.7443,   0.8373,   0.9303,   1.0234,
          1.1164,   1.2094,   1.3025,   1.3955,   1.4885,   1.5816,
          1.6746,   1.7676,   1.8607,   2.3258,   2.7910,   3.2562,
          3.7213,   4.1865,   4.6517,   5.1169,   5.5820,   6.0472,
          6.5124,   6.9775,   7.4427,   7.9079,   8.3730,   8.8382 ]

    v = [ 0.0000,  -0.0042,  -0.0164,  -0.0364,  -0.0640,  -0.0996, 
         -0.1440,  -0.1993,  -0.2718,  -0.3609,  -0.4613,  -0.5715, 
         -0.6911,  -0.8204,  -0.9599,  -1.1110,  -1.2673,  -1.4241, 
         -1.5812,  -1.7387,  -1.8965,  -2.6933,  -3.5034,  -4.3246, 
         -5.1572,  -5.9968,  -6.8440,  -7.6947,  -8.5526,  -9.4209,
        -10.3005, -11.1660, -12.0403, -12.9294, -13.8307, -14.7419 ]

    max = np.exp(u[-1])

    if x < 1:
        return 0
    if x > max:
        return 1

    return 1-np.exp(scipy.interpolate.interp1d(u, v)(np.log(x)))


from detc import RA,DEC,Detc,DetE,Resp

def area(st, psi = 0, ctj = 1, gmst = 0, cl = 0.9, N = 4, ap = True,\
         ra = None, dec = None, ap0 = None, detc0 = None):
    '''
    Sky localization area in deg^2 at the [cl] confidence level
    '''
    if isinstance(ap, list):
        f = ap
    elif ap:
        if ra is None:
            ra = RA
        if dec is None:
            dec = DEC
        f = apat(psi, ctj, gmst, N, ra=ra, dec=dec)
    else:
        f = []
        for i in np.arange(N):
            f.append(np.ones(len(RA)))

    st = copy(st)
    n  = len(f[0])
    w  = []
    u  = np.ones(n)
    ws = np.zeros(n)
    wx = np.zeros(n)
    wy = np.zeros(n)
    wz = np.zeros(n)
    xx = []
    yy = []
    zz = []

    for i in range(0, N):
        if detc0 is not None:
            x,y,z = detc0[i]
        elif ra == RA and dec == DEC:
            x,y,z = detc(i)
        else:
            x,y,z = detcc(i, ra, dec)

        xx.append(x)
        yy.append(y)
        zz.append(z)

        if isinstance(st[i], np.ndarray):
            s = st[i]*f[i]
        else:
            s = st[i]*u*f[i]
        w.append(1/s/s)

        ws += w[i]
        wx += w[i]*x
        wy += w[i]*y
        wz += w[i]*z

    c   = 299792458
    a0  = c*c*ws
    auu = np.zeros(n)
    avv = np.zeros(n)
    auv = np.zeros(n)
    aue = np.zeros(n)
    ave = np.zeros(n)
    aee = np.zeros(n)

    for i in range(0, N):
        d   = (ws/w[i]+1)*a0;
        wsx = wx-ws*xx[i]
        wsy = wy-ws*yy[i]
        wsz = wz-ws*zz[i]

        auu += wsx*wsx/d
        avv += wsy*wsy/d
        auv += wsx*wsy/d
        aue += wsx*wsz/d
        ave += wsy*wsz/d
        aee += wsz*wsz/d

    a  = auu
    b  = auv
    c  = avv
    d  = sqrt((a-c)*(a-c)+4*b*b)
    aa = 2/(a+c+d)
    bb = 2/(a+c-d)
    r  = sqrt(-2*log(1-cl));
    ca = 180*180/pi*r*r*sqrt(aa*bb);

    # Correction factor tuned with Bayestar
    cc = np.ones(len(ca))*0.6

    return ca/cc

def detc(i):
    '''
    Detector(HLVK, x,y,z) coordinates in m
    after rotating the GW source location to the North pole (0,0,1)
    '''
    n  = len(RA)
    u  = np.ones(n)
    R0 = 6378.1e3*u
    x  = R0*Detc[i*3]
    y  = R0*Detc[i*3+1]
    z  = R0*Detc[i*3+2]
    return x,y,z

def detcc(i, ra, dec, dete=DetE):
    n  = len(ra)
    ph = ra
    th = np.ones(n)*np.pi/2-dec
    cp = np.cos(ph)
    sp = np.sin(ph)
    ct = np.cos(th)
    st = np.sin(th)
    u  = np.ones(n)
    x0 = dete[i][0]
    y0 = dete[i][1]
    z0 = dete[i][2]
    R0 = 6378.1e3*u
    x1 = ct*cp*x0+ct*sp*y0-st*z0;
    y1 =   -sp*x0   +cp*y0;
    z1 = st*cp*x0+st*sp*y0+ct*z0;

    return x1*R0,y1*R0,z1*R0

from detc import Resp

def apat(psi = 0, ctj = 1, gmst = 0, N = 4, ra=None, dec=None):
    '''
    Detector antenna patterns at each (RA, DEC)
    '''
    if ra is None:
        ra = RA
    if dec is None:
        dec = DEC

    n    = len(ra)
    u    = np.ones(n)
    ph   = ra-u*gmst
    th   = u*pi/2-dec
    cosg = cos(-ph)
    sing = sin(-ph)
    cosd = cos(pi/2-th)
    sind = sin(pi/2-th)
    cosp = cos(psi)
    sinp = sin(psi)

    x = [ -cosp*sing -sinp*cosg*sind, -cosp*cosg +sinp*sing*sind, sinp*cosd ]
    y = [  sinp*sing -cosp*cosg*sind,  sinp*cosg +cosp*sing*sind, cosp*cosd ]
    f = []

    for i in range(0, N):
        fp = np.zeros(n)
        fc = np.zeros(n)

        rs = Resp[i]

        for j in range(0, 3):
            dx = u*rs[j*3]*x[0] +u*rs[j*3+1]*x[1] +u*rs[j*3+2]*x[2]
            dy = u*rs[j*3]*y[0] +u*rs[j*3+1]*y[1] +u*rs[j*3+2]*y[2]
            fp += x[j]*dx-y[j]*dy
            fc += x[j]*dy+y[j]*dx

        fp *= (1+ctj*ctj)/2
        fc *= ctj;

        f.append(1/sqrt(fp*fp+fc*fc))

    return f

from scipy import stats

def gen(psds, m=1.4, n= 1000000, thd= 12, thd2= 4, fl= 30, verb= 0,\
        ca= 0, dc= 1):
    '''
    Range estimation with MC approach
    '''
    fs,ps = loadPSDs(psds)

    zh = szhor(fs[0:1], ps[0:1], m=m, thd=thd2)
    vh = z2v(zh)
    dh = omg.LumD(zh)
    zz = np.linspace(0, 1, 21)*zh
    nd = len(fs)

    dd = []
    zz[0] = omg.z(1)
    for z in zz:
        dd.append(omg.LumD(z))

    snr = []
    sgt = []
    for i in np.arange(nd):
        snr.append(np.ndarray(0))
        sgt.append(np.ndarray(0))
        for (d,z) in zip(dd, zz):
            snr[i] = np.append(snr[i],\
                      SNR(fs[i], ps[i], d=d, m=m*(1+z), fl=fl)\
                      *d/(1+z)**(5./6))

    for (d,z) in zip(dd, zz):
        st = stz(fs, ps, z, m=m)
        for i in np.arange(nd):
            sgt[i] = np.append(sgt[i], st[i] /d*(1+z)**(5./6))

    if verb > 1:
        print 'dh=',dh,vh
        print 'SNR= ',snr

    x = np.linspace(0, 1, 100)*dh
    y = []
    x[0] = 1
    for xx in x:
        y.append(d2r(xx)**3)

    dst = scipy.interpolate.interp1d(y, x)(stats.uniform(0, y[-1]).rvs(n))
    z   = omg.zp(dst)
    ra  = stats.uniform(0, 2 * np.pi).rvs(n)
    dec = np.arcsin(stats.uniform(-1, 2).rvs(n))
    ctj = stats.uniform(-1, 2).rvs(n)
    psi = stats.uniform(0, 2*np.pi).rvs(n)
    ap  = apat(psi, ctj, ra=ra, dec=dec, N=nd)
    ss  = np.zeros(n)
    ns  = np.zeros(n, dtype=int)

    df  = []
    for i in np.arange(nd):
        df.append(stats.uniform(0, 1).rvs(n))

    for i in np.arange(nd):
        sn  = scipy.interpolate.interp1d(dd, snr[i])(dst)
        sn  = sn/ap[i]/dst*(1+z)**(5./6)
        ii  = np.where(df[i] > dc)[0]
        if len(ii):
            sn[ii] = 0

        ii  = np.where(sn > thd2)[0]
        ss += sn**2
        ns[ii] += 1

    ss[np.where(ns < 2)[0]] = 0
    ss  = np.sqrt(ss)
    ii  = np.where(ss > thd)[0]
    ns  = len(ii)
    rng = v2r(vh*ns/n)

    if verb > 0:
        rn0 = frange(psds, m=m, fl=fl, dc=dc)
        rr  = (rng/rn0-1)*100
        if ca == 0:
            print 'm= %5.1f range= %7.2f %7.2f (%6.3f %%)' % (m, rng, rn0, rr)

    if ca == 0:
        return rng

    # Sky localization CA range
    z   = z  [ii]
    dst = dst[ii]
    ra  = ra [ii]
    dec = dec[ii]
    ctj = ctj[ii]
    psi = psi[ii]
    st  = []
    ns  = np.ones(len(z), dtype=int)*nd
    for i in np.arange(nd):
        df[i] = df[i][ii]

    for i in np.arange(nd):
        st.append(np.ndarray(0))
        st[i]  = scipy.interpolate.interp1d(dd, sgt[i])(dst)
        st[i] *= dst/(1+z)**(5./6)
        ii     = np.where(df[i] > dc)[0]
        if len(ii):
            st[i][ii]  = 100
            ns   [ii] -= 1

    ii  = np.where(ns >= 3)[0]
    ra  = ra [ii]
    dec = dec[ii]
    ctj = ctj[ii]
    psi = psi[ii]
    for i in np.arange(nd):
        st[i] = st[i][ii]

    aa  = area(st, psi, ctj, N=len(st), ra=ra, dec=dec)

    if ca == -1:
        cdf = np.ndarray(0)
        cas = np.logspace(-1, 3, 100)
        for ca in cas:
            na  = len(np.where(aa < ca)[0])
            arn = v2r(vh*na/n)
            cdf = np.append(cdf, (arn/rng)**3)
        return cas, cdf

    else:
        na  = len(np.where(aa < ca)[0])
        arn = v2r(vh*na/n)

        if verb > 0:
            ar0 = frange(psds, m=m, opt='CA', thd=ca, dc=dc)
            rr  = (arn/ar0-1)*100
            print 'm= %5.1f range= %7.2f (%4.1f %%) '\
                                  '%7.2f (%6.3f %%) [%2.0f %%]'\
                % (m, arn, 100*(arn/rng)**3, ar0, rr, 100/np.sqrt(na))

    return arn
