
import numpy as np
from numpy import pi,sin,cos,sqrt,log,exp,power
from scipy import interpolate

def SNR(f, psd, d=100, m=1.4, apprx= None, fpw=0, fl=10, fh=1024, **args):
    """
    Get SNR for (f,psd) with waveform (apprx)
    """

    if apprx == None:
        apprx = 'TaylorF2' if m < 5 else 'IMRPhenomD'

    f,psd = frng(f, psd, fl, fh, fmax(m, apprx))

    if apprx == 'IMRPhenomD':
        return SNR_IMR(f, psd, d, m, fpw)
    else:
        return SNR_TF2(f, psd, d, m, fpw)


def fmax(m, apprx):
    """
    Get maximum frequency for waveform (apprx) with mass (m)
    """
    if apprx == 'IMRPhenomD':
        mt = m*2
        ms = 4.925491025543575903e-6  # Ms/c^3

        return 0.2/mt/ms              # Cutoff frequency

    else:  # TaylorF2
        Gc = 6.67384e-11
        C  = 299792458
        p6 = 14.6969384566990691      # 6^(-3./2)
        Ms = 1.98854695496146147e30   # Solar mass in kg

        return C**3/(p6*pi*Gc*2*m*Ms) # FISCO


def frng(f, psd, fl=30, fh=1024, fm=1024):
    i1 = np.where(f >= fl)[0][0]
    i2 = np.where(f <  fm)[0]
    i3 = np.where(f <  fh)[0][-1]
    i2 = i2[-1] if len(i2) > 0 else 0

    if i3 < i2:
        i2 = i3

    return f[i1:i2+1], psd[i1:i2+1]


def SNR_TF2(f, psd, d=100, m=1.4, fpw=0):
    """
    TaylorF2 waveform amplitude
    """

    Gc  = 6.67384e-11
    C   = 299792458
    Ms  = 1.98854695496146147e30             # Solar mass in kg
    Mpc = 3.08567758149136727891e22          # Mpc

    mc = power(m*m, 3./5)/pow(2*m, 1./5)*Ms  # chirp mass

    h  = power(f, -7./3+fpw*2)
    s  = np.trapz(h/psd, f)

    return 4*sqrt(s*5./96)*power(pi, -2./3)*C*power(Gc*mc/C**3, 5./6)/d/Mpc

def SNR_IMR(f, psd, d=100, m=30, fpw=0):
    """
    Parameterization of IMRPhenomD/P waveform amplitude
    assuming no spin (a1=a2=0) and same mass (m1=m2=m)
    """

    # Total mass
    mt = m*2
    # Ms/c^3
    ms = 4.925491025543575903411922162094833998e-6
    # Spline nodes
    xp = [ -10,    -9,   -8,    -7,   -6,    -5,   -4.7, -4.5, -4.4, -4.27, \
            -4.27, -4,   -3.7,  -3.5, -3.2,  -3,   -2.8, -2.7, -2.6, -2.554,\
            -2.55, -2.5, -2.45, -2.4, -2.35, -2.3, -2.2, -2.1, -2,   -1.9,  \
            -1.8, -1.7, -1.6 ]
    yp = [  6.795,   5.731,   4.539,  3.364,  2.177,\
            0.955,   0.572,   0.310,  0.177,  0.002,\
            0.002,  -0.360,  -0.743, -0.971, -1.243,\
           -1.372,  -1.477,  -1.536, -1.615, -1.663,\
           -1.667,  -1.751,  -1.913, -2.193, -2.602,\
           -3.118,  -4.312,  -5.564, -6.822, -8.086,\
           -9.375, -10.705, -12.091 ]

    f1 = interpolate.interp1d(xp[ 0:10], yp[ 0:10], kind="cubic")
    f2 = interpolate.interp1d(xp[10:20], yp[10:20], kind="cubic")
    f3 = interpolate.interp1d(xp[19:21], yp[19:21], kind="linear")
    f4 = interpolate.interp1d(xp[20:33], yp[20:33], kind="cubic")

    # Amplitude scaling factor from compnent mass 
    mm = 2*(np.log(m)-3.4)-50.5818
    lx = np.log(f*mt*ms)

    i1 = np.where(lx < xp[ 9]) # -4.27
    i2 = np.where(lx < xp[19]) # -2.554
    i3 = np.where(lx < xp[20]) # -2.55
    i4 = np.where(lx < xp[-1]) # -1.6

    i1 = i1[0][-1]+1 if len(i1[0])>0 else -1
    i2 = i2[0][-1]+1 if len(i2[0])>0 else -1
    i3 = i3[0][-1]+1 if len(i3[0])>0 else -1
    i4 = i4[0][-1]+1 if len(i4[0])>0 else -1

    y1 = f1(lx[ 0:i1])                              if i1 >= 0 else []
    y2 = f2(lx[i1:i2]) if i1 >= 0 else f2(lx[0:i2]) if i2 >= 0 else []
    y3 = f3(lx[i2:i3]) if i2 >= 0 else f3(lx[0:i3]) if i3 >= 0 else []
    y4 = f4(lx[i3:i4]) if i3 >= 0 else f4(lx[0:i4]) if i4 >= 0 else []

    y = np.append(np.append(np.append(y1, y2), y3), y4)
    h = np.exp(y+mm) *100/d

    return np.sqrt(4*np.trapz(h*h/psd*power(f, fpw*2), f));
