
import numpy as np
from scipy import integrate
from scipy import constants

class Omega:
    """
    LALCosmologyCalculator
    The set of functions in this module implements the standard
    cosmological distance measures
    defined a Friedmann-Robertson-Walker-Lemaitre metric.
    For a detailed reference to the various formulae, see Hogg 1999
    ( http://arxiv.org/abs/astro-ph/9905116 ).
    """

    def __init__(self):
        self.h = 0.679
        self.om = 0.3065
        self.ol = 0.6935
        self.ok = 1.0-self.om-self.ol
        self.w0 = -1.0
        self.w1 = 0.0
        self.w2 = 0.0

    # XLALHubbleDistance
    # Computes the Hubble distance, independent of the redshift.
    # This is the distance with sets the units for all others.
    # It returns the distance in Mpc.
    # Eq. 4 in Hogg 1999 ( http://arxiv.org/abs/astro-ph/9905116 )
    def HD(self):
        return 1e-3*constants.c/(100*self.h)

    # XLALHubbleParameter
    # Computes the inverse of the Hubble parameter at redshift z
    # Eq. 14 in Hogg 1999 ( http://arxiv.org/abs/astro-ph/9905116 )
    def Einv(self, z):
        x  = 1.0+z;
        om = self.om;
        ol = self.ol;
        ok = self.ok;
        w0 = self.w0;
        w1 = self.w1;
        w2 = self.w2;
        E  = np.sqrt(om*x*x*x+ok*x*x+ol*pow(x,3.*(1.0+w0+w1+w2))\
                        *np.exp(-3.0*((w1+w2)*z/x+w2*z*z/(2.0*x*x))))
        return 1/E; 

    # XLALIntegrateHubbleParameter
    # Computes the integral of inverse of the Hubble parameter at redshift z.
    def intEi(self, z):
        return integrate.quad(self.Einv, 0, z)[0]

    # XLALComovingLOSDistance
    # Computes the comoving line-of-sight distance
    # (usually called the proper distance) by integrating the Hubble parameter.
    # Eq. 15 in Hogg 1999 ( http://arxiv.org/abs/astro-ph/9905116 ).
    def LOSD(self, z):
        return self.HD()*self.intEi(z)

    # XLALComovingTransverseDistance
    # Computes the comoving transverse distance 
    # from the comoving line-of-sight distance (proper distance) and
    # Eq. 16 in Hogg 1999 ( http://arxiv.org/abs/astro-ph/9905116 )
    def TrvD(self, z):
        ok = self.ok
        dh = self.HD()
        dc = self.LOSD(z)
        if (abs(ok) < 1e-7):
            return dc
        if (ok > 1e-7):
            return dh*np.sinh(np.sqrt(ok)*dc/dh)/np.sqrt(ok)

        return dh*np.sin(np.sqrt(abs(ok))*dc/dh)/np.sqrt(abs(ok))

    # XLALLuminosityDistance
    # Computes the luminosity distance as the angular distance
    # divided by (1+z)^2
    # Eq. 21 in Hogg 1999 ( http://arxiv.org/abs/astro-ph/9905116 )
    def LumD(self, z):
        return self.TrvD(z)*(1+z)

    # XLALUniformComovingVolumeDensity
    # This function computes the value of a uniform probability density 
    # distribution over the comoving volume at redshift z.
    def CVD(self, z):
        return self.CVE(z)/(1+z)

    # XLALComovingVolumeElement
    # This function computes the comoving volume element (a 4&pi shell) 
    # between redshift z and z+dz.
    def CVE(self, z):
        dm = self.TrvD(z)
        E  = self.Einv(z)
        return 4*np.pi*dm*dm*E*self.HD()


OMEGA = Omega()
"""
# These are default values
OMEGA.h = 0.679
OMEGA.om = 0.3065
OMEGA.ol = 0.6935
OMEGA.ok = 1.0 - OMEGA.om - OMEGA.ol
OMEGA.w0 = -1.0
OMEGA.w1 = 0.0
OMEGA.w2 = 0.0
"""
