from __future__ import division
import copy
from collections import OrderedDict
import numpy as np
import scipy
from scipy import constants

import lal
import lalsimulation

from .cosmology import OMEGA

##################################################

# default inspiral waveform parameters
# face-on 1.4/1.4 Msolar inspiral at 100 Mpc distance
DEFAULT_PARAMS = OrderedDict([
    ('approximant', None),
    ('distance', 100e6),
    ('m1', 1.4),
    ('m2', 1.4),
    ('S1x', 0.0),
    ('S1y', 0.0),
    ('S1z', 0.0),
    ('S2x', 0.0),
    ('S2y', 0.0),
    ('S2z', 0.0),
    ('inclination', 0.0),
    ('f_ref', 0.0),
    ('phiRef', 0.0),
    ('longAscNodes', 0.0),
    ('eccentricity', 0.0),
    ('meanPerAno', 0.0),
    ('deltaF', None),
    ('f_min', None),
    ('f_max', None),
    ('LALpars', None),
    ])

DEFAULT_APPROXIMANT_BNS = 'TaylorF2'
DEFAULT_APPROXIMANT_BBH = 'IMRPhenomD'

##################################################

def _get_waveform_params(**kwargs):
    params = OrderedDict(DEFAULT_PARAMS)
    params.update(**kwargs)
    # use waveform approximant appropriate to type
    if not params['approximant']:
        if params['m1'] >= 5 and params['m2'] >= 5:
            params['approximant'] = DEFAULT_APPROXIMANT_BBH
        else:
            params['approximant'] = DEFAULT_APPROXIMANT_BNS
    return params


def gen_waveform(omega=OMEGA, **params):
    """Generate frequency-domain inspiral waveform

    Returns a tuple of (freq, h_tilde^plus, h_tilde^cross).

    The waveform is generated with
    lalsimulation.SimInspiralChooseFDWaveform().  Keyword arguments
    are used to update the default waveform parameters (1.4/1.4
    Msolar, optimally-oriented, 100 Mpc, (see DEFAULT_PARAMS macro)).
    The mass parameters ('m1' and 'm2') should be specified in solar
    masses and the 'distance' parameter should be specified in
    parsecs**.  Waveform approximants may be given as string names
    (see `lalsimulation` documentation for more info).  If the
    approximant is not specified explicitly, DEFAULT_APPROXIMANT_BNS
    waveform will be used if either mass is less than 5 Msolar and
    DEFAULT_APPROXIMANT_BBH waveform will be used otherwise.

    For example, to generate a 20/20 Msolar BBH waveform:

    >>> hp,hc = waveform.gen_waveform('m1'=20, 'm2'=20)

    **NOTE: The requirement that masses are specified in solar masses
    and distances are specified in parsecs is different than that of
    the underlying lalsimulation method which expects mass and
    distance parameters to be in SI units.

    """
    iparams = _get_waveform_params(**params)

    # convert to SI units
    iparams['distance'] *= lal.PC_SI
    iparams['m1'] *= lal.MSUN_SI
    iparams['m2'] *= lal.MSUN_SI
    iparams['approximant'] = lalsimulation.SimInspiralGetApproximantFromString(iparams['approximant'])

    m = iparams['m1'] + iparams['m2']

    # calculate delta F based on frequency of inner-most stable
    # circular orbit ("fisco")
    fisco = (constants.c**3)/(constants.G*(6**1.5)*2*np.pi*m)
    df = 2**(np.max([np.floor(np.log(fisco/4096)/np.log(2)), -6]))

    # FIXME: are these limits reasonable?
    iparams['deltaF'] = df
    iparams['f_min'] = 0.1
    iparams['f_max'] = 10000

    hp, hc = lalsimulation.SimInspiralChooseFDWaveform(**iparams)

    freq = hp.f0 + np.arange(len(hp.data.data)) * hp.deltaF

    return freq, hp.data.data, hc.data.data

##################################################

class CBCWaveform:
    """CBC waveform class for inspiral range calculations

    Generates waveform and stores parameters for later
    transformations.

    NOTE: Only the amplitude of the plus polarization of the generated
    waveform is stored, as that's all that's needed for calculating
    SNR.

    """
    def __init__(self, freq=None, z0=0.1, omega=OMEGA, **params):
        """Initialize waveform

        Waveform is generated with gen_waveform, and the parameters of
        generation, frequency array, and amplitude of h_plus waveform
        component are stored.

        If a frequency vector is provided the waveform will be
        interpolated to those points, which can speed up subsequenty
        z-shifted SNR() calculations considerably.  Otherwise the PSD
        for SNR() calculations must be specified at the frequencies in
        self.freq.

        """
        self.z0 = z0
        self.omega = omega
        self.params = _get_waveform_params(**params)
        self.d0 = lal.LuminosityDistance(self.omega, self.z0) * 1e6
        iparams = copy.deepcopy(self.params)
        iparams['distance'] = self.d0
        iparams['m1'] *= 1.0 + self.z0
        iparams['m2'] *= 1.0 + self.z0
        self.freq, hp, hc = gen_waveform(omega=omega, **iparams)
        self.h = np.absolute(hp)
        if freq is not None:
            self.h = scipy.interpolate.interp1d(self.freq, self.h)(freq)
            self.freq = freq


    def z_scale(self, z):
        """Scale/shift waveform amplitude and frequency arrays for a given redshift

        Returns transformed (freq, h) tuple.

        """
        d = lal.LuminosityDistance(self.omega, z) * 1e6
        fscale = (1+self.z0)/(1+z)
        hscale = self.d0/d * ((1+z)/(1+self.z0))**2
        return self.freq*fscale, self.h*hscale


    def SNR(self, psd, z=0):
        """Calculate waveform SNR against given PSD

        If redshift z provided, SNR will be calculated for a waveform
        transformed to that redshift.

        PSD must be specified at the same frequency points as
        used/generated at initialization.

        """
        if z != 0:
            f, h = self.z_scale(z)

            #h = scipy.interpolate.interp1d(f, h, bounds_error=False, fill_value=(h[0], 0))(self.freq)
            # Modified by SH for python2.7
            h0 = h[0]
            h = scipy.interpolate.interp1d(f, h, bounds_error=False)(self.freq)
            ii = np.where(h!=h)[0]
            if len(ii) > 0:
                h[ii] = h0 if ii[0] == 0 else 0
            # Modified by SH for python2.7
        else:
            h = self.h

        return np.sqrt(4*np.trapz((h**2)/psd, self.freq))
